@@ -1,219 +1,53 @@
 
 
-import chunk
 import typing
 import collections
-import math
-import codecs
-#import datetime
 
-from ._Hex import _Hex
-from ._Bits import _Bits
+from .RawValue import RawValue
+from .DumperSettings import DumperSettings
+#from ._Hex import _Hex
+#from ._Bits import _Bits
+from ._ConverterFunctions import _ConverterFunctions as _CF
+from ._ByteChunker import _ByteChunker
 
 
 
 
 
 
-class DumperSettings(object):
 
-	def __init__(self):
-		self.showPrimitivesWithType = False
-		self.showDictKeysWithType = False
-		self.showComplexStructsWithType = False
-		self.compactSequenceLengthLimit = 8
-		self.compactSequenceItemLengthLimit = 50
-		self.bytesLineSize = 32
-		self.compactBytesLinesLengthLimit = 32
-	#
-
-#
 
 DEFAULT_DUMPER_SETTINGS = DumperSettings()
 
 
 
 
-def _str_shortenText(text:str) -> str:
-	if len(text) > 40:
-		return text[:40] + "..."
-	else:
-		return text
-#
-
-def _float_roundTo7FractionDigits(data:typing.Union[int,float]) -> float:
-	return round(data, 7)
-#
-
-def _float_roundTo6FractionDigits(data:typing.Union[int,float]) -> float:
-	return round(data, 6)
-#
-
-def _float_roundTo5FractionDigits(data:typing.Union[int,float]) -> float:
-	return round(data, 5)
-#
-
-def _float_roundTo4FractionDigits(data:typing.Union[int,float]) -> float:
-	return round(data, 4)
-#
-
-def _float_roundTo3FractionDigits(data:typing.Union[int,float]) -> float:
-	return round(data, 3)
-#
-
-def _float_roundTo2FractionDigits(data:typing.Union[int,float]) -> float:
-	return round(data, 2)
-#
-
-def _float_roundTo1FractionDigits(data:typing.Union[int,float]) -> float:
-	return round(data, 1)
-#
-
-def _int_toHex(data:int) -> typing.Union[int,str]:
-	if data < 0:
-		return data
-	return _Hex(data)
-#
-
-def _int_toBits(data:int) -> typing.Union[int,str]:
-	if data < 0:
-		return data
-	return _Bits(data)
-#
-
-def _byteChunker(data:bytes, chunkSize:int) -> typing.Sequence[bytes]:
-	assert isinstance(data, bytes)
-	assert isinstance(chunkSize, int)
-	assert chunkSize > 0
-
-	iFrom = 0
-	iTo = chunkSize
-	while iFrom < len(data):
-		yield data[iFrom:iTo]
-		iFrom = iTo
-		iTo += chunkSize
-#
-
-def _x_byteChunkToHex(data:bytes) -> str:
-	hexSpanLength = 16		# == 8 bytes for a span
-
-	s = codecs.encode(data, "hex").decode("ascii")
-	chunks = [ s[i:i+hexSpanLength] for i in range(0, len(s), hexSpanLength) ]
-	return " ".join(chunks)
-#
-
-def _x_byteChunkToASCII(data:bytes) -> str:
-	spanLength = 8		# == 8 bytes for a span
-
-	ret = []
-	for i, b in enumerate(data):
-		if (i % spanLength) == 0:
-			ret.append(" ")
-		if 32 <= b <= 127:
-			ret.append(chr(b))
-		else:
-			ret.append(".")
-
-	return "".join(ret)
-#
-
-#
-# Returns chunks of the specified data.
-#
-def _byteChunkerWithOfs(settings:DumperSettings, data:bytes, processorName:str = None) -> typing.Sequence[typing.Tuple[str,str,str]]:
-	assert isinstance(settings, DumperSettings)
-	assert isinstance(data, bytes)
-	chunkSize = settings.bytesLineSize
-	assert isinstance(chunkSize, int)
-	assert chunkSize > 0
-	hexStrPadded = chunkSize*2 + math.ceil(chunkSize / 8) - 1
-	if processorName is not None:
-		assert isinstance(processorName, str)
-
-	# ----
-
-	nTotalLength = len(data)
-	nTotalLines = math.ceil(nTotalLength / chunkSize)
-	formatStrFragment = None
-	formatStrFragmentEllipsis = None
-	if nTotalLength <= 256*256:
-		formatStrFragment = "{:04x}"
-		formatStrFragmentEllipsis = "... "
-	elif nTotalLength <= 256*256*256:
-		formatStrFragment = "{:06x}"
-		formatStrFragmentEllipsis = "...   "
-	else:
-		formatStrFragment = "{:08x}"
-		formatStrFragmentEllipsis = "...     "
-
-	# ----
-
-	skipFrom = -1
-	skipTo = -1
-	if processorName:
-		if processorName == "shorten":
-			skipFrom = settings.compactBytesLinesLengthLimit
-			skipTo = nTotalLines - 4
-			if skipFrom >= skipTo:
-				skipFrom = -1
-				skipTo = -1
-		else:
-			raise Exception("No such postprocessor: " + repr(processorName))
-
-	# ----
-
-	if skipFrom < 0:
-		# direct loop, no addtional if statements
-		iFrom = 0
-		iTo = chunkSize
-		while iFrom < len(data):
-			chunk = data[iFrom:iTo]
-			yield formatStrFragment.format(iFrom), _x_byteChunkToHex(chunk).ljust(hexStrPadded), _x_byteChunkToASCII(chunk)
-			iFrom = iTo
-			iTo += chunkSize
-	else:
-		# loop with if statements
-		lineNo = 0
-		iFrom = 0
-		iTo = chunkSize
-		while iFrom < len(data):
-			chunk = data[iFrom:iTo]
-			if skipFrom <= lineNo <= skipTo:
-				if skipFrom == lineNo:
-					yield formatStrFragmentEllipsis, "...".ljust(hexStrPadded), "..."
-			else:
-				yield formatStrFragment.format(iFrom), _x_byteChunkToHex(chunk).ljust(hexStrPadded), _x_byteChunkToASCII(chunk)
-			iFrom = iTo
-			iTo += chunkSize
-			lineNo += 1
-#
 
 
+_PRIMITIVE_POST_PROCESSORS = {
+	# map associating processor name to tuple consisting of type (or type list) and callable
 
+	"shorten": (str, _CF.str_shortenText),
+	"hex": (int, _CF.int_toHex),
+	"bit": (int, _CF.int_toBits),
+	"round1": ((float, int), _CF.float_roundTo1FractionDigits),
+	"round2": ((float, int), _CF.float_roundTo2FractionDigits),
+	"round3": ((float, int), _CF.float_roundTo3FractionDigits),
+	"round4": ((float, int), _CF.float_roundTo4FractionDigits),
+	"round5": ((float, int), _CF.float_roundTo5FractionDigits),
+	"round6": ((float, int), _CF.float_roundTo6FractionDigits),
+	"round7": ((float, int), _CF.float_roundTo7FractionDigits),
 
-
-POST_PROCESSORS = {
-	"shorten": (str, _str_shortenText),
-	"hex": (int, _int_toHex),
-	"bit": (int, _int_toBits),
-	"round1": ((float, int), _float_roundTo1FractionDigits),
-	"round2": ((float, int), _float_roundTo2FractionDigits),
-	"round3": ((float, int), _float_roundTo3FractionDigits),
-	"round4": ((float, int), _float_roundTo4FractionDigits),
-	"round5": ((float, int), _float_roundTo5FractionDigits),
-	"round6": ((float, int), _float_roundTo6FractionDigits),
-	"round7": ((float, int), _float_roundTo7FractionDigits),
-
-	"str_shorten": (str, _str_shortenText),
-	"float_round7": ((float, int), _float_roundTo7FractionDigits),
-	"float_round6": ((float, int), _float_roundTo6FractionDigits),
-	"float_round5": ((float, int), _float_roundTo5FractionDigits),
-	"float_round4": ((float, int), _float_roundTo4FractionDigits),
-	"float_round3": ((float, int), _float_roundTo3FractionDigits),
-	"float_round2": ((float, int), _float_roundTo2FractionDigits),
-	"float_round1": ((float, int), _float_roundTo1FractionDigits),
-	"int_hex": (int, _int_toHex),
-	"int_bit": (int, _int_toBits),
+	"str_shorten": (str, _CF.str_shortenText),
+	"float_round7": ((float, int), _CF.float_roundTo7FractionDigits),
+	"float_round6": ((float, int), _CF.float_roundTo6FractionDigits),
+	"float_round5": ((float, int), _CF.float_roundTo5FractionDigits),
+	"float_round4": ((float, int), _CF.float_roundTo4FractionDigits),
+	"float_round3": ((float, int), _CF.float_roundTo3FractionDigits),
+	"float_round2": ((float, int), _CF.float_roundTo2FractionDigits),
+	"float_round1": ((float, int), _CF.float_roundTo1FractionDigits),
+	"int_hex": (int, _CF.int_toHex),
+	"int_bit": (int, _CF.int_toBits),
 }
 
 
@@ -223,23 +57,17 @@ class _Omitted:
 	pass
 #
 
-class RawValue:
-
-	def __init__(self, text:str) -> None:
-		self.text = text
-	#
-
-#
-
 _OMITTED = _Omitted()
 
 
 
 
 
+DumpCtx = typing.NewType("DumpCtx", object)
+
 class DumpCtx(object):
 
-	_TYPE_MAP = {}				# type -> function
+	_TYPE_MAP:typing.Dict[type,typing.Callable[[DumpCtx,str,typing.Any,typing.Union[str,None]],None]] = {}				# type -> function
 
 	def __init__(self, s:DumperSettings, outputLines:list, exitAppend:str, prefix:str):
 		self.__s = s
@@ -260,7 +88,7 @@ class DumpCtx(object):
 	#
 
 	def dumpVarRaw(self, varName:str, value:RawValue) -> None:
-		self._dumpX(varName + " = ", value.text)
+		self._dumpX(varName + " = ", value)
 	#
 
 	#
@@ -305,9 +133,13 @@ class DumpCtx(object):
 	################################################################################################################################
 
 	#
-	# This method outputs a value (recursively).
+	# This method outputs a value (recursively). It is the main dump method.
 	# To achieve this this method analyses the data type of the specified value and invokes individual type processing methods if available.
 	#
+	# @param		str extraPrefix			(required) A prefix to use
+	# @param		any value				(optional) The value to print
+	# @param		str? processorName		(optional) A value output processor
+	#
 	def _dumpX(self, extraPrefix:str, value, processorName:str = None):
 		if value is None:
 			self._dumpPrimitive(extraPrefix, None, processorName)
@@ -332,10 +164,10 @@ class DumpCtx(object):
 		# is it an object with a DumpMixin?
 
 		if isinstance(value, DumpMixin):
-			self._dumpObj(extraPrefix, value)
+			self._dumpObj(extraPrefix, value, processorName)
 			return
 
-		# is it derived from on of our types?
+		# is it derived from one of our types?
 
 		for storedT, m in DumpCtx._TYPE_MAP.items():
 			if isinstance(value, storedT):
@@ -362,24 +194,37 @@ class DumpCtx(object):
 	#
 	# Dump the specified object.
 	#
+	# @param	str processorName		(optional) The name of an output processor.
+	#									Supports: "shorten"
+	#
 	def _dumpObj(self, extraPrefix:str, value:object, processorName:str = None):
-		self.outputLines.append(self.prefix + extraPrefix + "<" + value.__class__.__name__ + "(")
-
-		ctx = DumpCtx(self.__s, self.outputLines, None, self.prefix + "\t")
-		with ctx as ctx2:
-			if hasattr(value, "_dump"):
-				value._dump(ctx2)
-			elif hasattr(value, "_dumpVarNames"):
-				ctx2.dumpVars(value)
+		if processorName == "shorten":
+			if hasattr(value, "_dumpShort"):
+				value._dumpShort(ctx2)
 			else:
-				raise Exception("Improper object encountered for prettyprinting: " + type(value).__name__)
+				self.outputLines.append(self.prefix + extraPrefix + "<" + value.__class__.__name__ + "(...)>")
 
-		self.outputLines.append(self.prefix + ")>")
+		else:
+			self.outputLines.append(self.prefix + extraPrefix + "<" + value.__class__.__name__ + "(")
+
+			ctx = DumpCtx(self.__s, self.outputLines, None, self.prefix + "\t")
+			with ctx as ctx2:
+				if hasattr(value, "_dump"):
+					value._dump(ctx2)
+				elif hasattr(value, "_dumpVarNames"):
+					ctx2.dumpVars(value)
+				else:
+					raise Exception("Improper object encountered for prettyprinting: " + type(value).__name__)
+
+			self.outputLines.append(self.prefix + ")>")
 	#
 
 	#
 	# Dump the specified dictionary.
 	#
+	# @param		str processorName			(optional) The processor name. This name is passed to recursive calls of _dumpX() so that it is applied
+	#											to every value. Additionally if "shorten" is specified the dictionary itself will be shortened.
+	#
 	def _dumpDict(self, extraPrefix:str, value:dict, processorName:str = None):
 		e = (type(value).__name__ + ":") if self.__s.showComplexStructsWithType else ""
 
@@ -387,11 +232,17 @@ class DumpCtx(object):
 
 		ctx = DumpCtx(self.__s, self.outputLines, None, self.prefix + "\t")
 		with ctx as ctx2:
+			i = 0
 			for k, v in value.items():
+				if processorName == "shorten":
+					if i >= 3:
+						ctx2._dumpRawValue("", RawValue("..."))
+						break
 				if processorName == "omitValues":
 					v = _OMITTED
 				ctx2._dumpX(self._dictKeyToStr(k) + " : ", v)
 				self.outputLines[-1] += ","
+				i += 1
 
 		self.outputLines.append(self.prefix + "}")
 	#
@@ -413,20 +264,30 @@ class DumpCtx(object):
 	#
 	# Dump the specified list.
 	#
+	# @param		str processorName			(optional) The processor name. This name is passed to recursive calls of _dumpX() so that it is applied
+	#											to every value. Additionally if "shorten" is specified the list itself will be shortened.
+	#
 	def _dumpList(self, extraPrefix:str, value:list, processorName:str = None):
 		e = (type(value).__name__ + ":") if self.__s.showComplexStructsWithType else ""
 
 		if self._canCompactSequence(value):
-			self.outputLines.append(self.prefix + extraPrefix + e + "[ " + self._compactSequence(value, processorName) + " ]")
+			rawText = self._compactSequence(value, processorName)
+			self.outputLines.append(self.prefix + extraPrefix + e + "[ " + rawText + " ]")
 
 		else:
 			self.outputLines.append(self.prefix + extraPrefix + e + "[")
 
 			ctx = DumpCtx(self.__s, self.outputLines, None, self.prefix + "\t")
 			with ctx as ctx2:
+				i = 0
 				for vItem in value:
+					if processorName == "shorten":
+						if i >= 3:
+							ctx2._dumpRawValue("", RawValue("..."))
+							break
 					ctx2._dumpX("", vItem, processorName)
 					self.outputLines[-1] += ","
+					i += 1
 
 			self.outputLines.append(self.prefix + "]")
 	#
@@ -443,7 +304,7 @@ class DumpCtx(object):
 		else:
 			self.outputLines.append(self.prefix + extraPrefix + e + "<")
 
-			for sOfs, chunk, sAscii in _byteChunkerWithOfs(self.__s, value, processorName):
+			for sOfs, chunk, sAscii in _ByteChunker.chunkWithOfs(self.__s, value, processorName):
 				self.outputLines.append(self.prefix + "\t" + sOfs + "  " + chunk + "  " + sAscii)
 
 			if len(value) == 1:
@@ -457,20 +318,30 @@ class DumpCtx(object):
 	#
 	# Dump the specified tuple.
 	#
+	# @param		str processorName			(optional) The processor name. This name is passed to recursive calls of _dumpX() so that it is applied
+	#											to every value. Additionally if "shorten" is specified the list itself will be shortened.
+	#
 	def _dumpTuple(self, extraPrefix:str, value:set, processorName:str = None):
 		e = (type(value).__name__ + ":") if self.__s.showComplexStructsWithType else ""
 
 		if self._canCompactSequence(value):
-			self.outputLines.append(self.prefix + extraPrefix + e + "( " + self._compactSequence(value, processorName) + " )")
+			rawText = self._compactSequence(value, processorName)
+			self.outputLines.append(self.prefix + extraPrefix + e + "( " + rawText + " )")
 
 		else:
 			self.outputLines.append(self.prefix + extraPrefix + e + "(")
 
 			ctx = DumpCtx(self.__s, self.outputLines, None, self.prefix + "\t")
 			with ctx as ctx2:
+				i = 0
 				for vItem in value:
+					if processorName == "shorten":
+						if i >= 3:
+							ctx2._dumpRawValue("", RawValue("..."))
+							break
 					ctx2._dumpX("", vItem, processorName)
 					self.outputLines[-1] += ","
+					i += 1
 
 			self.outputLines.append(self.prefix + ")")
 	#
@@ -526,7 +397,18 @@ class DumpCtx(object):
 	#
 
 	def _dumpRawValue(self, extraPrefix:str, value:RawValue):
-		self.outputLines.append(self.prefix + extraPrefix + value.text)
+		if isinstance(value.textOrLines, str):
+			self.outputLines.append(self.prefix + extraPrefix + value.textOrLines)
+			return
+
+		firstLine = value.textOrLines[0]
+		moreLines = value.textOrLines[1:]
+
+		self.outputLines.append(self.prefix + extraPrefix + firstLine)
+
+		_prefix2 = self.prefix + " " * len(extraPrefix)
+		for line in moreLines:
+			self.outputLines.append(_prefix2 + line)
 	#
 
 	def _dumpOmitted(self, extraPrefix:str, value, processorName:str = None):
@@ -550,6 +432,7 @@ class DumpCtx(object):
 		return True
 	#
 
+	# TODO: support "shorten" to shorten the list
 	def _compactSequence(self, someSequence, processorName:str = None) -> str:
 		ret = []
 		for v in someSequence:
@@ -586,9 +469,9 @@ class DumpCtx(object):
 		else:
 			# process value before converting it to str
 			if processorName:
-				if processorName not in POST_PROCESSORS:
+				if processorName not in _PRIMITIVE_POST_PROCESSORS:
 					raise Exception("No such postprocessor: " + repr(processorName))
-				postProcessorTypeCompatibility, postProcessor = POST_PROCESSORS[processorName]
+				postProcessorTypeCompatibility, postProcessor = _PRIMITIVE_POST_PROCESSORS[processorName]
 				if isinstance(value, postProcessorTypeCompatibility):
 					value = postProcessor(value)
 
@@ -689,42 +572,72 @@ class Dumper(object):
 
 
 
+################################################################################################################################
+################################################################################################################################
+"""
+
+Dumpable objects should be defined like this:
+
+	class FooBar(SomeBaseClassA,SomeMixinB,jk_prettyprintobj.DumpMixin)
+		....
+	#
+
+Dumpable objects should then implement one of these two methods:
+
+	def _dump(self, ctx:jk_prettyprintobj.DumpCtx):
+		ctx.dumpVar(...)
+	#
+
+or:
+
+	def _dumpVarNames(self) -> typing.List[str]:
+		return [
+			"....",
+		]
+	#
+
+and maybe additionally:
+
+	def _dumpShort(self, ctx:jk_prettyprintobj.DumpCtx):
+		ctx.dumpVar(...)
+	#
+
+"""
+################################################################################################################################
+################################################################################################################################
+
+
+
+
 class DumpMixin:
 
 	__slots__ = tuple()
 
-	def dump(self, prefix:str = None, printFunc = None) -> None:
+	def __dump(self, prefix:str = None) -> Dumper:
 		dumper = Dumper()
 		with dumper.createContext(self, prefix) as dumper2:
 			if not dumper2._isDumpableObj(self):
 				raise Exception("Improper object encountered for prettyprinting: " + self.__class__.__name__ + " - Either implement _dump(ctx:DumpCtx) or _dumpVarNames()!")
 			dumper2._dumpObj("", self)
+		return dumper
+	#
+
+	def dump(self, prefix:str = None, printFunc = None) -> None:
+		dumper = self.__dump(prefix)
 		dumper.print(printFunc)
 	#
 
 	def dumpToStr(self, prefix:str = None) -> str:
-		dumper = Dumper()
-		with dumper.createContext(self, prefix) as dumper2:
-			if not dumper2._isDumpableObj(self):
-				raise Exception("Improper object encountered for prettyprinting: " + self.__class__.__name__ + " - Either implement _dump(ctx:DumpCtx) or _dumpVarNames()!")
-			dumper2._dumpObj("", self)
+		dumper = self.__dump(prefix)
 		return dumper.toStr()
 	#
 
-	################################################################
-	# NOTE: Dumpable objects must implement one of these methods:
+	def dumpToFile(self, filePath:str) -> None:
+		dumper = self.__dump()
+		assert isinstance(filePath, str)
+		with open(filePath, "w", newline="\n", encoding="UTF-8") as fout:
+			fout.write(dumper.toStr())
 	#
-	#	def _dump(self, ctx:DumpCtx):
-	#		ctx.dumpVar(...)
-	#	#
-	#
-	#	def _dumpVarNames(self) -> typing.List[str]:
-	#		return [
-	#			"....",
-	#		]
-	#	#
-	#
-	################################################################
 
 #
 
